@@ -1,6 +1,5 @@
 /* enchant
  * Copyright (C) 2003-2004 Joan Moratinos <jmo@softcatala.org>, Dom Lachowicz
- * Copyright (C) 2016-2021 Reuben Thomas <rrt@sc3d.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,8 +16,8 @@
  * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  * Boston, MA 02110-1301, USA.
  *
- * In addition, as a special exception, the copyright holders
- * give permission to link the code of this program with
+ * In addition, as a special exception, Dom Lachowicz
+ * gives permission to link the code of this program with
  * non-LGPL Spelling Provider libraries (eg: a MSFT Office
  * spell checker backend) and distribute linked combinations including
  * the two.  You must obey the GNU General Public License in all
@@ -30,44 +29,33 @@
 
 /*
  * This is the Hunspell Enchant Backend.
- * Hunspell is by László Németh. See: https://hunspell.github.io/
+ * Hunspell is by László Németh. See: http://hunspell.github.io/
  */
 
 #include "config.h"
 
 #include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
+#include <string.h> 
 
 #include <string>
 #include <vector>
 
 #include "enchant-provider.h"
+#include "unused-parameter.h"
 
 #include <hunspell/hunspell.hxx>
-// hunspell itself uses this definition (which only supports the BMP)
-#define MAXWORDUTF8LEN (MAXWORDLEN * 3)
+
+/* Some versions of hunspell (1.4.x) don't have this defined. */
+/* This is the defined value at that point */
+#ifndef MAXWORDLEN
+#define MAXWORDLEN 176
+#endif
 
 #include <glib.h>
 
 /***************************************************************************/
 
-static const char *empty_string = "";
-
-static char *do_iconv(GIConv conv, const char *word) {
-	// g_iconv() does not declare its 'in' parameter const, but iconv() does.
-	char *in = const_cast<char *>(word);
-	size_t len_in = strlen(in);
-	size_t len_out = len_in * 3;
-	char *out_buf = g_new0(char, len_out + 1);
-	char *out = out_buf;
-	size_t result = g_iconv(conv, &in, &len_in, &out, &len_out);
-	if (static_cast<size_t>(-1) == result)
-		return nullptr;
-	*out = '\0';
-	return out_buf;
-}
-
 class HunspellChecker
 {
 public:
@@ -77,7 +65,6 @@ public:
 	bool checkWord (const char *word, size_t len);
 	char **suggestWord (const char* const word, size_t len, size_t *out_n_suggs);
 	const char *getWordchars ();
-	bool apostropheIsWordChar;
 
 	bool requestDictionary (const char * szLang);
 
@@ -85,7 +72,6 @@ private:
 	GIConv  m_translate_in; /* Selected translation from/to Unicode */
 	GIConv  m_translate_out;
 	Hunspell *hunspell;
-	char *wordchars; /* Value returned by getWordChars() */
 };
 
 /***************************************************************************/
@@ -97,7 +83,7 @@ g_iconv_is_valid(GIConv i)
 }
 
 HunspellChecker::HunspellChecker()
-: apostropheIsWordChar(false), m_translate_in(nullptr), m_translate_out(nullptr), hunspell(nullptr), wordchars(nullptr)
+: m_translate_in(nullptr), m_translate_out(nullptr), hunspell(nullptr)
 {
 }
 
@@ -108,61 +94,86 @@ HunspellChecker::~HunspellChecker()
 		g_iconv_close(m_translate_in);
 	if (g_iconv_is_valid(m_translate_out))
 		g_iconv_close(m_translate_out);
-	free(wordchars);
 }
 
 bool
 HunspellChecker::checkWord(const char *utf8Word, size_t len)
 {
-	if (len > MAXWORDUTF8LEN || !g_iconv_is_valid(m_translate_in))
+	if (len > MAXWORDLEN || !g_iconv_is_valid(m_translate_in))
 		return false;
 
 	// the 8bit encodings use precomposed forms
 	char *normalizedWord = g_utf8_normalize (utf8Word, len, G_NORMALIZE_NFC);
-	char *out = do_iconv(m_translate_in, normalizedWord);
+	char *in = normalizedWord;
+	char word8[MAXWORDLEN + 1];
+	char *out = word8;
+	size_t len_in = strlen(in);
+	size_t len_out = sizeof( word8 ) - 1;
+	size_t result = g_iconv(m_translate_in, &in, &len_in, &out, &len_out);
 	g_free(normalizedWord);
-	if (out == NULL)
+	if (static_cast<size_t>(-1) == result)
+		return false;
+	*out = '\0';
+	if (hunspell->spell(word8))
+		return true;
+	else
 		return false;
-	bool result = hunspell->spell(std::string(out)) != 0;
-	free(out);
-	return result;
 }
 
 char**
 HunspellChecker::suggestWord(const char* const utf8Word, size_t len, size_t *nsug)
 {
-	if (len > MAXWORDUTF8LEN
+	if (len > MAXWORDLEN 
 		|| !g_iconv_is_valid(m_translate_in)
 		|| !g_iconv_is_valid(m_translate_out))
 		return nullptr;
 
 	// the 8bit encodings use precomposed forms
 	char *normalizedWord = g_utf8_normalize (utf8Word, len, G_NORMALIZE_NFC);
-	char *out = do_iconv(m_translate_in, normalizedWord);
+	char *in = normalizedWord;
+	char word8[MAXWORDLEN + 1];
+	char *out = word8;
+	size_t len_in = strlen(in);
+	size_t len_out = sizeof(word8) - 1;
+	size_t result = g_iconv(m_translate_in, &in, &len_in, &out, &len_out);
 	g_free(normalizedWord);
-	if (out == NULL)
+	if (static_cast<size_t>(-1) == result)
 		return nullptr;
 
-	std::vector<std::string> sugMS = hunspell->suggest(out);
-	g_free(out);
-	*nsug = sugMS.size();
+	*out = '\0';
+	char **sugMS;
+	*nsug = hunspell->suggest(&sugMS, word8);
 	if (*nsug > 0) {
 		char **sug = g_new0 (char *, *nsug + 1);
-		for (size_t i=0, j=0; i<*nsug; i++) {
-			const char *in = sugMS[i].c_str();
-			out = do_iconv(m_translate_out, in);
-			if (out != NULL)
-				sug[j++] = out;
+		for (size_t i=0; i<*nsug; i++) {
+			in = sugMS[i];
+			len_in = strlen(in);
+			len_out = MAXWORDLEN;
+			char *word = g_new0(char, len_out + 1);
+			out = word;
+			if (static_cast<size_t>(-1) == g_iconv(m_translate_out, &in, &len_in, &out, &len_out)) {
+				for (size_t j = i; j < *nsug; j++)
+					free(sugMS[j]);
+				free(sugMS);
+
+				*nsug = i;
+				return sug;
+			}
+			*(out) = 0;
+			sug[i] = word;
+			free(sugMS[i]);
 		}
+		free(sugMS);
 		return sug;
 	}
-	return nullptr;
+	else
+		return nullptr;
 }
 
-_GL_ATTRIBUTE_PURE const char*
+const char*
 HunspellChecker::getWordchars()
 {
-	return static_cast<const char *>(wordchars);
+	return hunspell->get_wordchars();
 }
 
 static void
@@ -170,8 +181,9 @@ s_buildDictionaryDirs (std::vector<std::string> & dirs)
 {
 	dirs.clear ();
 
+	gchar * tmp;
 	char * config_dir = enchant_get_user_config_dir ();
-	gchar * tmp = g_build_filename (config_dir, "hunspell", nullptr);
+	tmp = g_build_filename (config_dir, "hunspell", nullptr);
 	dirs.push_back (tmp);
 	free (config_dir);
 	g_free(tmp);
@@ -189,7 +201,7 @@ s_buildDictionaryDirs (std::vector<std::string> & dirs)
 		{
 			tmp = g_build_filename(enchant_prefix, "share", "enchant", "hunspell", nullptr);
 			dirs.push_back (tmp);
-			free(enchant_prefix);
+			g_free(enchant_prefix);
 			g_free(tmp);
 		}
 
@@ -241,16 +253,16 @@ static bool is_plausible_dict_for_tag(const char *dir_entry, const char *tag)
     size_t tag_len = strlen(tag);
 
     if (dir_entry_len - dic_suffix_len < tag_len)
-	return false;
+        return false;
     if (strcmp(dir_entry+dir_entry_len-dic_suffix_len, dic_suffix) != 0)
-	return false;
+        return false;
     if (strncmp (dir_entry, tag, tag_len) != 0)
-	return false;
+        return false;
     //e.g. requested dict for "fi",
     //reject "fil_PH.dic"
     //allow "fi-FOO.dic", "fi_FOO.dic", "fi.dic", etc.
     if (!ispunct(dir_entry[tag_len]))
-	return false;
+        return false;
     return true;
 }
 
@@ -267,7 +279,7 @@ hunspell_request_dictionary (const char * tag)
 			return strdup (names[i].c_str());
 		}
 	}
-
+	
 	std::vector<std::string> dirs;
 	s_buildDictionaryDirs (dirs);
 
@@ -277,7 +289,7 @@ hunspell_request_dictionary (const char * tag)
 			const char *dir_entry;
 			while ((dir_entry = g_dir_read_name (dir)) != NULL) {
 				if (is_plausible_dict_for_tag(dir_entry, tag)) {
-					char *dict = g_build_filename (dirs[i].c_str(),
+					char *dict = g_build_filename (dirs[i].c_str(), 
 								       dir_entry, nullptr);
 					if(s_fileExists(s_correspondingAffFile(dict))) {
 						g_dir_close (dir);
@@ -304,11 +316,6 @@ HunspellChecker::requestDictionary(const char *szLang)
 	std::string aff(s_correspondingAffFile(dic));
 	if (s_fileExists(aff))
 	{
-		if (hunspell) {
-			delete hunspell;
-			free(wordchars);
-			wordchars = NULL;
-		}
 		hunspell = new Hunspell(aff.c_str(), dic);
 	}
 	free(dic);
@@ -320,14 +327,6 @@ HunspellChecker::requestDictionary(const char *szLang)
 	m_translate_in = g_iconv_open(enc, "UTF-8");
 	m_translate_out = g_iconv_open("UTF-8", enc);
 
-	wordchars = do_iconv(m_translate_out, hunspell->get_wordchars());
-	if (wordchars == NULL)
-		wordchars = strdup(empty_string);
-	if (wordchars == NULL)
-		return false;
-	apostropheIsWordChar = g_utf8_strchr(wordchars, -1, g_utf8_get_char("'")) ||
-		g_utf8_strchr(wordchars, -1, g_utf8_get_char("’"));
-
 	return true;
 }
 
@@ -347,10 +346,10 @@ static int
 hunspell_dict_check (EnchantDict * me, const char *const word, size_t len)
 {
 	HunspellChecker * checker = static_cast<HunspellChecker *>(me->user_data);
-
+	
 	if (checker->checkWord(word, len))
 		return 0;
-
+	
 	return 1;
 }
 
@@ -364,12 +363,8 @@ hunspell_dict_get_extra_word_characters (EnchantDict *me)
 static int
 hunspell_dict_is_word_character (EnchantDict *me, uint32_t uc, size_t n)
 {
+	(void)n;
 	HunspellChecker * checker = static_cast<HunspellChecker *>(me->user_data);
-	/* Accept quote marks anywhere except at the end of a word, as per
-	   hunspell's textparser.cxx/TextParser::next_token */
-	if ((uc == g_utf8_get_char("'") || uc == g_utf8_get_char("’")) && checker->apostropheIsWordChar) {
-		return n < 2;
-	}
 	return g_unichar_isalpha(uc) || g_utf8_strchr(checker->getWordchars(), -1, uc);
 }
 
@@ -409,8 +404,8 @@ hunspell_provider_enum_dicts (const char * const directory,
 
 extern "C" {
 
-static char **
-hunspell_provider_list_dicts (EnchantProvider * me _GL_UNUSED,
+static char ** 
+hunspell_provider_list_dicts (EnchantProvider * me _GL_UNUSED_PARAMETER, 
 			      size_t * out_n_dicts)
 {
 	std::vector<std::string> dict_dirs, dicts;
@@ -435,18 +430,18 @@ hunspell_provider_list_dicts (EnchantProvider * me _GL_UNUSED,
 }
 
 static EnchantDict *
-hunspell_provider_request_dict(EnchantProvider * me _GL_UNUSED, const char *const tag)
+hunspell_provider_request_dict(EnchantProvider * me _GL_UNUSED_PARAMETER, const char *const tag)
 {
 	HunspellChecker * checker = new HunspellChecker();
-
+	
 	if (!checker)
 		return NULL;
-
+	
 	if (!checker->requestDictionary(tag)) {
 		delete checker;
 		return NULL;
 	}
-
+	
 	EnchantDict *dict = g_new0(EnchantDict, 1);
 	dict->user_data = (void *) checker;
 	dict->check = hunspell_dict_check;
@@ -454,21 +449,21 @@ hunspell_provider_request_dict(EnchantProvider * me _GL_UNUSED, const char *cons
 	// don't implement personal, session
 	dict->get_extra_word_characters = hunspell_dict_get_extra_word_characters;
 	dict->is_word_character = hunspell_dict_is_word_character;
-
+	
 	return dict;
 }
 
 static void
-hunspell_provider_dispose_dict (EnchantProvider * me _GL_UNUSED, EnchantDict * dict)
+hunspell_provider_dispose_dict (EnchantProvider * me _GL_UNUSED_PARAMETER, EnchantDict * dict)
 {
 	HunspellChecker *checker = (HunspellChecker *) dict->user_data;
 	delete checker;
-
+	
 	g_free (dict);
 }
 
 static int
-hunspell_provider_dictionary_exists (struct str_enchant_provider * me _GL_UNUSED,
+hunspell_provider_dictionary_exists (struct str_enchant_provider * me _GL_UNUSED_PARAMETER,
 				     const char *const tag)
 {
 	std::vector <std::string> names;
@@ -491,13 +486,13 @@ hunspell_provider_dispose (EnchantProvider * me)
 }
 
 static const char *
-hunspell_provider_identify (EnchantProvider * me _GL_UNUSED)
+hunspell_provider_identify (EnchantProvider * me _GL_UNUSED_PARAMETER)
 {
 	return "hunspell";
 }
 
 static const char *
-hunspell_provider_describe (EnchantProvider * me _GL_UNUSED)
+hunspell_provider_describe (EnchantProvider * me _GL_UNUSED_PARAMETER)
 {
 	return "Hunspell Provider";
 }
